#! /usr/bin/perl

# (C) 2006 Pavel Mihaylov
#          Ben Waldron (additional code)

use utf8;
use strict;
use warnings;
no warnings 'uninitialized';

use Encode qw(encode decode);
use Fcntl;
use IO::Socket;
use IO::Handle;
use POSIX qw(getcwd);

use Glib qw(TRUE FALSE);
use Gtk2 -init;
use Gtk2::Gdk::Keysyms;
use IPC::Run qw(harness start pump finish timeout);

use constant RS => "\x1E";
use constant US => "\x1F";

use constant VERSION => "0.91";


# some defaults
my %defaults = (
    font        => 'Sans 13', # not used
    sharePath   => $ENV{TROLLET_SHARE} || '/usr/share/trollet',
    lispBin     => $ENV{TROLLET_LISP} || 'lkb',
    lkbSrc      => $ENV{TROLLET_LKBSRC} ,
);

my $lkb_h = harness [ $defaults{lispBin} ], '<pipe', \*LKB_IN, '>pipe', \*LKB_OUT;

# the main menu
my @mainmenu = (
    [ '_Grammar', [
        [ '_Load grammar...', { activate => sub { send_to_lisp('(read-script-file)') } } ],
        [ '_Reload current grammar...', { activate => sub { send_to_lisp('(reload-script-file)') } } ],
        undef,
#        [ '_Load Bulgarian grammar...', { activate => sub { send_to_lisp('(read-script-file-aux "/home/avataar/LKB/bulgarian/lkb/script")') } } ],
#        undef,
        [ '_Options', { activate => \&show_options_dialog } ],
        undef,
        [ '_Quit', { activate => \&quit_trollet } ]
    ] ],
    [ '_View', [
        [ 'Type _hierarchy...', { activate => sub { send_to_lisp('(show-type-tree)') } } ],
        [ 'Type _definition...', { activate => sub { send_to_lisp('(show-type-spec)') } } ],
        [ '_Expanded type...', { activate => sub { send_to_lisp('(show-type)') } } ],
        [ '_Lex entry...', { activate => sub { send_to_lisp('(show-lex)') } } ],
        [ '_Word entries...', { activate => sub { send_to_lisp('(show-words)') } } ],
        [ '_Other entry...', { activate => sub { send_to_lisp('(show-other)') } } ],
        [ '_Grammar rule...', { activate => sub { send_to_lisp('(show-grammar-rule)') } } ],
        [ 'Le_xical rule...', { activate => sub { send_to_lisp('(show-lex-rule)') } } ],
        [ '_All words', { activate => sub { send_to_lisp('(display-lex-words)') } } ],
    ] ],
    [ '_Parse', [
        [ '_Parse input...', { activate => sub { send_to_lisp('(do-parse)') } } ],
        # not enabled by default in LKB
        #[ 'Parse _file...', { activate => sub { send_to_lisp('(parse-with-preprocessor)') } } ],
        [ '_Redisplay parse', { activate => sub { send_to_lisp('(show-parse)') } } ],
        [ '_Show parse chart', { activate => sub { send_to_lisp('(show-chart)') } } ],
        [ '_Batch parse...', { activate => sub { send_to_lisp('(parse-sentences)') } } ],
        [ '_Compare...', { activate => sub { send_to_lisp('(compare-parses)') } } ]
    ] ],
    [ '_MRS', [
        [ '_Load munger...', { activate => sub { send_to_lisp('(read-mrs-rule-file)') } } ],
        [ '_Clear munger', { activate => sub { send_to_lisp('(clear-mrs-rules)') } } ],
        [ 'Toggle _base output', { activate => sub { send_to_lisp('(toggle-mrs-base)') } } ],
        [ 'Toggle _scoping', { activate => sub { send_to_lisp('(toggle-mrs-scoping)') } } ],
        [ 'Toggle _RMRS', { activate => sub { send_to_lisp('(toggle-rmrs)') } } ],
        [ 'Toggle RMRS (_XML)', { activate => sub { send_to_lisp('(toggle-rmrs-xml)') } } ],
        [ 'R_MRS Compare', { activate => sub { send_to_lisp('(compare-rmrs-interactive)') } } ],
    ] ],
    [ 'Ge_nerate', [
        [ 'Redisplay _realisation', { activate => sub { send_to_lisp('(show-gen-result)') } } ],
        [ 'Redisplay _MRS', { activate => sub { send_to_lisp('(show-generator-input)') } } ],
        [ 'Show _gen chart', { activate => sub { send_to_lisp('(show-gen-chart)') } } ],
        [ '_Load heuristics...', { activate => sub { send_to_lisp('(read-gen-rule-file)') } } ],
        [ '_Clear heuristics', { activate => sub { send_to_lisp('(clear-gen-rules)') } } ],
        [ '_Index', { activate => sub { send_to_lisp('(index-for-generator)') } } ],
        [ '_Start server', { activate => sub { send_to_lisp('(start-generator-server)') } } ],
        [ 'Sto_p server', { activate => sub { send_to_lisp('(stop-generator-server)') } } ],
    ] ],
    [ '_Debug', [
        [ '_Check lexicon', { activate => sub { send_to_lisp('(batch-check-lexicon)') } } ],
        [ "_Find features' type...", { activate => sub { send_to_lisp('(find-type-from-features)') } } ],
        [ 'Print _parser chart', { activate => sub { send_to_lisp('(print-chart-toplevel)') } } ],
        [ 'Print _token chart', { activate => sub { send_to_lisp('(print-token-chart-toplevel)') } } ],
        [ 'Print _generator chart', { activate => sub { send_to_lisp('(print-gen-chart-toplevel)') } } ],
        [ 'Print _lexical rule FSM', { activate => sub { send_to_lisp('(print-lrfsm-toplevel)') } } ],

    ] ],
    [ '_Advanced', [
	    [ '_LexDb', [
            [ "_Load ('rev' entries)", { activate => sub { send_to_lisp('(command-merge-into-lexdb)') } } ],
	        [ '_Dump', { activate => sub { send_to_lisp('(command-dump-lexdb)') } } ],
	        [ '_Export (TDL file)', { activate => sub { send_to_lisp('(command-export-lexicon-to-tdl)') } } ],
	        [ 'Set _Filter', { activate => sub { send_to_lisp('(command-set-filter-lexdb)') } } ],
	        [ '_Import (TDL file)', { activate => sub { send_to_lisp('(command-load-tdl-to-scratch)') } } ],
	        [ "Di_splay private 'rev'", { activate => sub { send_to_lisp('(command-show-private-rev)') } } ],
	        [ "_Commit private 'rev'", { activate => sub { send_to_lisp('(command-commit-private-rev)') } } ],
	        [ "C_lear private 'rev'", { activate => sub { send_to_lisp('(command-clear-private-rev)') } } ],
	        [ "I_ndex new 'lex' entries", { activate => sub { send_to_lisp('(command-index-new-lex-entries)') } } ]
	   ] ],
       [ 'LKB _Options', { activate => sub { send_to_lisp('(get-parameters)') } } ],
       [ '_Tidy up', { activate => sub { send_to_lisp('(clear-non-parents)') } } ],
       [ '_Create quick check file...', { activate => sub { send_to_lisp('(interactive-create-check-paths)') } } ]
    ] ],
    [ '_Help', [
        [ '_About', { activate => \&show_about_dialog } ]
    ] ]
);


my %dialogs;

my @history;
my $history_idx = 0;
my $restore_text;
my $restore_pos;
my $last_char = "\n";

my $trolletdir = "$ENV{HOME}/.trollet";
my $trolletsocket = "$trolletdir/socket";

# make the dir
unless (-d $trolletdir) {
    unlink $trolletdir;
    mkdir $trolletdir;
}
unlink $trolletsocket;

# a window to hold the program
my $window = Gtk2::Window->new('toplevel');
$window->set_default_size(600, 400);
$window->signal_connect(delete_event => \&quit_trollet);
$window->set_default_icon_list(
    Gtk2::Gdk::Pixbuf->new_from_file("$defaults{sharePath}/images/trollet-24x24.png"),
    Gtk2::Gdk::Pixbuf->new_from_file("$defaults{sharePath}/images/trollet-32x32.png"),
    Gtk2::Gdk::Pixbuf->new_from_file("$defaults{sharePath}/images/trollet-48x48.png"),
);

# a vbox to align things nicely
my $vbox = Gtk2::VBox->new(0, 0);

# the main menu
my $menubar = Gtk2::MenuBar->new;
make_menus($menubar, \@mainmenu);

my $textbuffer = Gtk2::TextBuffer->new;

my $cmd_tag = $textbuffer->create_tag(undef, "foreground", "blue");
my $parse_tag = $textbuffer->create_tag(undef, "foreground", "dark green");
my $error_tag = $textbuffer->create_tag(undef, "foreground", "red");

my $textview = Gtk2::TextView->new_with_buffer($textbuffer);
$textview->set_editable(0);
$textview->can_focus(0);
$textview->set_cursor_visible(0);
$textview->set_wrap_mode('word');

# set up scrolling
my $scrolledwindow = Gtk2::ScrolledWindow->new;
$scrolledwindow->set_policy('never', 'always');
$scrolledwindow->add($textview);

# text entry + parse button
my $hbox2 = Gtk2::HBox->new;
my $entry = Gtk2::Entry->new;
my $parse = Gtk2::Button->new('LISP');

$hbox2->pack_start($entry, TRUE, TRUE, 0);
$hbox2->pack_start($parse, FALSE, FALSE, 0);

$entry->signal_connect(activate => \&process_entry);
$entry->signal_connect(key_press_event => \&process_history );
$parse->signal_connect(clicked => \&lisp_entry);

# status bar
my $statusbar = Gtk2::Statusbar->new;

# add everything to the vbox
$vbox->pack_start($menubar, FALSE, FALSE, 0);
$vbox->pack_start($scrolledwindow, TRUE, TRUE, 0);
$vbox->pack_start($hbox2, FALSE, FALSE, 0);
$vbox->pack_start($statusbar, FALSE, FALSE, 0);

# add the vbox to the window and show it
$window->add($vbox);
$entry->grab_focus;
$window->show_all;

# start LKB (with LUI and no LKB Top)
$ENV{LUI} = 'on';
$ENV{LKB_GUI_EXTERNAL} = 't';
$lkb_h->start;

autoflush LKB_IN 1;
binmode LKB_IN, ':utf8';
binmode LKB_OUT, ':utf8';

# load LKB from source if necessary
if ($defaults{lkbSrc}) {
    send_to_lisp(qq{;; Loading LKB from source
(unless (find-package :lkb)
 (require :climxm)
 (load "$defaults{lkbSrc}/general/loadup")
 (pushnew :lkb *features*))
(unless (find-package :lkb)
 (load-system "mrs")
 (in-package :lkb))}, 1);
}
send_to_lisp(qq{(load "$defaults{sharePath}/trollet.lisp")}, 1);



# what happens when we close the window
$window->signal_connect(delete_event => sub { Gtk2->main_quit; FALSE });

# snoop on a UNIX socket
my $server = IO::Socket::UNIX->new(Type => SOCK_STREAM, Local => $trolletsocket,
                                     Listen    => 10);

Glib::IO->add_watch($server->fileno, 'G_IO_IN', \&process_snooped);
Glib::IO->add_watch(fileno(LKB_OUT), 'G_IO_IN', \&process_lkb);
#Glib::Timeout->add(50, \&process_lkb);

# off we go
Gtk2->main;

######################### subs down here ###############################

# parse the text in the text entry
sub parse_entry {
    my $text = $entry->get_text;
    if ($text =~ /[^\s]/) {
        append_to_textview("Parsing ‘$text’\n", $parse_tag);
        $text =~ s/"/\\"/g;
        send_to_lisp(qq{(do-parse-tty "$text")});
        $entry->set_text('');
        add_history('parse', $text);
    }
}

# process the text entry as a LISP statement
sub lisp_entry {
    my $text = $entry->get_text;
    if ($text =~ /[^\s]/) {
        append_to_textview("> $text\n", $cmd_tag);
        send_to_lisp($text);
        $entry->set_text('');
        add_history('lisp', $text);
        $entry->grab_focus;
    }
}

# process the text entry when the user presses enter
sub process_entry {
    my $text = $entry->get_text;
    if ($text =~ /^\s*[\w\d]/) {
        parse_entry;
    } elsif ($text =~ /[^\s]/) {
        lisp_entry;
    }
}

# process history in the main text entry
sub process_history {
    my ($widget, $event) = @_;
    if ($event->keyval == $Gtk2::Gdk::Keysyms{Up} or
        $event->keyval == $Gtk2::Gdk::Keysyms{KP_Up}) {
        if ($history_idx > 0) {
            if ($history_idx == @history) {
                $restore_text = $widget->get_text;
                $restore_pos = $widget->get_position;
            }
            $widget->set_text($history[--$history_idx]->[1]);
            $widget->set_position(-1);
        }
        return TRUE;
    } elsif ($event->keyval == $Gtk2::Gdk::Keysyms{Down} or
             $event->keyval == $Gtk2::Gdk::Keysyms{KP_Down}) {
        if ($history_idx+1 < @history) {
            $widget->set_text($history[++$history_idx]->[1]);
            $widget->set_position(-1);
        } elsif ($history_idx+1 == @history) {
            $widget->set_text($restore_text);
            $widget->set_position($restore_pos);
            $history_idx++;
        }
        return TRUE;
    }
    return FALSE;
}

sub add_history {
    my ($type, $text) = @_;
    unless (@history and $history[-1]->[0] eq $type and
                         $history[-1]->[1] eq $text) {
        push @history, [ $type, $text ];
    }
    $history_idx = @history;
    #print "added: $type/$text, $history_idx\n";
}

# show a file dialog and load a grammar script
sub load_grammar {
    my $filename = show_file_dialog('Load a grammar script', 'open', 'grammar');
    $filename =~ s/"/\\"/g;
    send_to_lisp(qq{(read-script-file-aux "$filename")});
}

sub make_file_dialog {
    my $mode = shift;
    my $dialog = Gtk2::FileChooserDialog->new('Choose a file:', $window, $mode,
                'gtk-cancel', 'cancel', "gtk-$mode", 'ok');
    $dialog->set_do_overwrite_confirmation(TRUE);
    $dialog->set_current_folder(getcwd);
    return $dialog;
}

sub show_file_dialog {
    my ($title, $mode) = @_;
    unless (defined $dialogs{"file_sel_$title"}) {
        $dialogs{"file_sel_$title"} = make_file_dialog($mode);
    }
    $dialogs{"file_sel_$title"}->set_title($title);
    $dialogs{"file_sel_$title"}->show_all;
    my $filename = undef;
    if ($dialogs{"file_sel_$title"}->run eq 'ok') {
        $filename = $dialogs{"file_sel_$title"}->get_filename;
    }
    $dialogs{"file_sel_$title"}->hide;
    return $filename;
}


# send to LKB through LISP
sub send_to_lisp {
    my ($data, $no_error_handling) = @_;
    my $lkb_in;
    if ($no_error_handling) {
        $lkb_in = "$data\n";
    } else {
        $lkb_in = "(abort-on-error $data)\n";
    }
    print LKB_IN $lkb_in;
#    print STDERR "E: $lkb_in";
}


# build the menus
sub make_menus {
    my ($parent, $menus) = @_;
    foreach my $menu (@$menus) {
        my $mi = defined $menu ?
                    Gtk2::TearoffMenuItem->new_with_mnemonic($menu->[0]) :
                    Gtk2::SeparatorMenuItem->new;
        $parent->add($mi);
        if (ref($menu->[1]) eq 'HASH') {
            # signals
            $mi->signal_connect(%{$menu->[1]});
        } elsif (ref($menu->[1]) eq 'ARRAY') {
            # submenu
            my $m = Gtk2::Menu->new;
            $mi->set_submenu($m);
            make_menus($m, $menu->[1]);
        }
    }
}

sub view_lex {
    show_simple_dialog('view_lex', 'View a lex entry', 'Lex-id:',
                        '(show-lex-tr0ndar "%s")');
}

sub view_word {
    show_simple_dialog('view_word', 'View a word entry', 'Word:',
                        '(show-word-aux-tty "%s" t)');
}

sub view_exp_type {
    show_simple_dialog('view_exp_type', 'View expanded type', 'Type:',
                        '(show-lex-tr0ndar "%s")');
}

sub show_simple_dialog {
    my ($name, $title, $label, $lisp_template) = @_;

    unless (defined $dialogs{$name}) {
        $dialogs{$name} = make_simple_dialog($title, $label);
        $dialogs{$name}->show_all;
    }

    my $entry = get_simple_dialog_entry($dialogs{$name});
    my $old_text = $entry->get_text;

    $entry->select_region(0, -1);
    $entry->grab_focus;

    if ($dialogs{$name}->run eq 'ok') {
        $entry->set_text(my $text = uc $entry->get_text);
        if ($text ne '') {
            $text =~ s/"/\\"/g;
            $lisp_template =~ s/%s/$text/;
            send_to_lisp($lisp_template);
        }
    } else {
        $entry->set_text($old_text);
    }

    $dialogs{$name}->hide;
}

sub make_simple_dialog {
    my ($title, $msg) = @_;
    my $dialog = Gtk2::Dialog->new($title, $window,
                                [qw/modal destroy-with-parent/],
                                'gtk-ok'     => 'ok',
                                'gtk-cancel' => 'reject');
    $dialog->set_default_response('ok');
    
    my $label = Gtk2::Label->new($msg);
    $label->set_alignment(0, 0.5);
    
    my $entry = Gtk2::Entry->new;
    $entry->set_activates_default(TRUE);
    
    $dialog->vbox->pack_start($label, FALSE, FALSE, 0);
    $dialog->vbox->pack_start($entry, TRUE, TRUE, 0);

    return $dialog;
}

sub get_simple_dialog_entry {
    my $dialog = shift;
    my (undef, $entry) = $dialog->vbox->get_children;
    return $entry;
}

sub show_about_dialog {
    unless (defined $dialogs{about}) {
        $dialogs{about} = Gtk2::AboutDialog->new;
        $dialogs{about}->set_authors('Pavel Mihaylov <bin@bash.info>',
                                     'Ben Waldron <bmw20@hf.ntnu.no>');
        $dialogs{about}->set_license(<<HERE
TROndheim LingLab Engineering Tool

Developed at LingLab\@NTNU, Trondheim

This code is free software; you can redistribute it and/or modify it under the terms of either:

    a) the Artistic Licence, version 2.0beta14 or later,
    b) the GNU General Public Licence, version 2 or later.

HERE
        );
        $dialogs{about}->set_name('Trollet');
        $dialogs{about}->set_comments('The better way to use LKB');
        $dialogs{about}->set_version(VERSION);
        $dialogs{about}->set_copyright('Copyright © 2006 Pavel Mihaylov');
        #$dialogs{about}->set_wrap_license(TRUE); # needs Gtk2 2.8
    }
    $dialogs{about}->run;
    $dialogs{about}->hide;
}

sub show_options_dialog {
    show_message_dialog('info', 'ok', "Not implemented");
}

sub show_message_dialog {
    my $dialog = Gtk2::MessageDialog->new($window, 'destroy-with-parent', @_);
    my $response = $dialog->run;
    $dialog->destroy;
    return $response;
}

sub show_strings_movable {
    my ($title, $keys, $vals, $width, $choices) = @_;
    my $dialog = Gtk2::Dialog->new($title, $window,
                              [qw/modal destroy-with-parent/],
                              'gtk-ok'     => 'ok',
                              'gtk-cancel' => 'reject');
    $dialog->set_default_response('ok');
    my $vbox = $dialog->vbox;
    my $table = Gtk2::Table->new($#{$keys}+1, 2, FALSE);
    for(0..$#{$keys}) {
        #print "item: $keys->[$_] -> $vals->[$_]\n";
        my $widget;
        if ($vals->[$_] eq ":CHECK-BOX") {
            my $checkbutton = Gtk2::CheckButton->new($keys->[$_]);
            $table->attach_defaults($checkbutton, 0, 2, $_, $_+1);
            
        } elsif ($_ == 0 and @$choices) { # choices
                #print "choices given!\n";
                my $liststore = Gtk2::ListStore->new('Glib::String');
                my $selected;
                foreach(0..$#{$choices}) {
                    my $iter = $liststore->append;
                    $selected = $_
                        if $choices->[$_] eq $vals->[0];
                    $liststore->set($iter, 0, $choices->[$_]);
                }
                #print "sel: $selected\n";

                my $treeview = Gtk2::TreeView->new($liststore);
                $treeview->set_enable_search(TRUE);
                $treeview->signal_connect('row-activated' =>
                    sub {
                        my $treeview = shift;
                        $dialog->activate_default;
                    });

                my $renderer = Gtk2::CellRendererText->new;
                $keys->[$_] =~ s/^\s*//;
                my $column = Gtk2::TreeViewColumn->new_with_attributes($keys->[$_],
                    $renderer, text => 0);
                
                $treeview->append_column($column);
                
                if (defined $selected) {
                    my $path = Gtk2::TreePath->new($selected);
                    $treeview->scroll_to_cell($path, undef, TRUE, 0.5, undef);
                    $treeview->set_cursor($path, undef, FALSE);
                }

                my $scrolled = Gtk2::ScrolledWindow->new($treeview->get_hadjustment,
                    $treeview->get_vadjustment);
                $scrolled->add($treeview);
                $scrolled->set_size_request(-1, 200);

                $table->attach_defaults($scrolled, 0, 2, $_, $_+1);
                
        } else {
            my $entry = Gtk2::Entry->new;
            if ($width ne 'NIL') {
                $entry->set_width_chars($width / 7);
            }
            $entry->set_activates_default(TRUE);
            $entry->set_text($vals->[$_]);
            $table->attach_defaults($entry, 1, 2, $_, $_+1);
            $keys->[$_] =~ s/^\s*//;
            my $label = Gtk2::Label->new($keys->[$_]);
            #$label->set_alignment(0, 0.5);
            $table->attach_defaults($label, 0, 1, $_, $_+1);
        }
        #$vbox->add($widget);
    }
    $vbox->add($table);
    $dialog->show_all;
    my $response = $dialog->run;
    if ($response eq 'ok') {
        my @children = reverse $table->get_children;
        my @result;
        foreach my $child (@children) {
            #print "$child\n";
            if ($child =~ /Gtk2::Entry/) {
                push @result, $child->get_text;
            }
            if ($child =~ /Gtk2::CheckButton/) {
                push @result, $child->get_active?'T':'NIL';
            }
            if ($child =~ /Gtk2::ScrolledWindow/) {
                my $treeview = $child->get_child;
                my $treeselection = $treeview->get_selection;
                my ($path) = $treeselection->get_selected_rows;
                $path = $path->to_string;
                #print "path: $path\n";
                push @result, $choices->[$path];
                
            }
        }
        $response = join(" ", map { s/"/\\"/g; qq{"$_"} } @result);
    } else {
        $response = "";
    }
    $dialog->destroy;
    #print "($response)\n";
    return "($response)";
}

sub process_snooped {
    my $sock = $server->accept;
    binmode $sock, ':utf8';

    #print "snooping on $sock\n";

    $/ = undef;
    my $data = <$sock>;
    my $result = 'NIL';

    #print STDERR "snooped: $data\n";

    my ($cmd, @args) = split RS, $data;
    if ($cmd eq 'message') {
        show_message_dialog('info', 'ok', $args[0]);
    } elsif ($cmd eq 'existing file') {
        my $filename;
        do {
            $filename = show_file_dialog($args[0], 'open');
            if (defined $filename) {
                unless (-f $filename) {
                    show_message_dialog('error', 'ok',
                        'File does not exist or is not readable.');
                    undef $filename;
                }
            } else {
                $filename = 'NIL';
            }
        } until defined $filename;
        $result = $filename;
    } elsif ($cmd eq 'new file') {
        my $filename;
        do {
            $filename = show_file_dialog($args[0], 'save');
            if (defined $filename) {
                if ($filename eq $args[1]) {
                    show_message_dialog('error', 'ok',
                        'Output file cannot be the same as input file.');
                    undef $filename;
                }
#                if (-f $filename) {
#                    my $resp = show_message_dialog('question', 'yes-no',
#                        'File exists, do you want to overwrite it?');
#                    if ($resp ne 'yes') {
#                        undef $filename;
#                    }
#                }
            } else {
                $filename = 'NIL';
            }
        } until defined $filename;
        $result = $filename;
    } elsif ($cmd eq 'strings movable') {
        my $title = $args[0];
        my @keys = split US, $args[1];
        my @vals = split US, $args[2];
        my $width = $args[3];
        my @choices = split US, $args[4];
        #print "keys: ", join(";", @keys), "\n";
        #print "vals: ", join(";", @vals), "\n";
        #print "width: $width";
        #print "choices: ", join(";", @choices), "\n";
        $result = show_strings_movable($title, \@keys, \@vals, $width, \@choices);
    }
    unless ($result =~ /^(\(|NIL$|T$)/) {
        print $sock qq{"$result"\n};
    } else {
        print $sock "$result\n";
    }
    $sock->close;
    #print "snooping done\n";
    return TRUE;
}


sub process_lkb {
    my $lkb_out;
#    print STDERR "read LKB input...";
    sysread LKB_OUT, $lkb_out, 10000;
#    print STDERR " ok\n";
#    print STDERR "text buffer updated: $lkb_out";
    $lkb_out =~ s/^LKB\(\d+\): //mg;
    $lkb_out =~ s/^(T|NIL|-?\d+)(\n|\z)//mg; # rip off anything that looks like
                                          # T, NIL or a number on a line of its own
#    $lkb_out = "\n$lkb_out" unless substr($lkb_out,-1) eq "\n";
    append_to_textview($lkb_out);
    return TRUE;
}

sub append_to_textview {
    my ($data, $tag) = @_;
    
    if ($data =~ /\S/s) {
        # prepend a new line if there isn't one at the end of the already appended text
        $data = "\n$data" unless $last_char eq "\n" or substr($data, 0, 1) eq "\n";
        if (defined $tag) {
            # show everything in a colour given by $tag
            $textbuffer->insert_with_tags($textbuffer->get_end_iter, $data, $tag);
            $last_char = substr $data, -1;
        } else {
            # find the first error message
            my ($start, $error, $end) = $data =~ m{(.*?)<error>(.+?)</error>\n(.*)}s;
            if (defined $error) {
                # before the error
                $textbuffer->insert($textbuffer->get_end_iter, $start);
                # the error in a different colour
                $textbuffer->insert_with_tags($textbuffer->get_end_iter,
                    "ERROR: $error\n", $error_tag);
                $last_char = "\n";
                # do the rest and more errors if necessary
                append_to_textview($end);
            } else {
                # no errors, all in default colour
                $textbuffer->insert($textbuffer->get_end_iter, $data);
                $last_char = substr $data, -1;
            }
        }
        # scroll the textview to the end
        my $mark = $textbuffer->create_mark(undef, $textbuffer->get_end_iter, 1);
        $textview->scroll_to_mark($mark, 0, FALSE, 0, 1);
        $textbuffer->delete_mark($mark);
    }
}

# quits the program
sub quit_trollet {
    $lkb_h->kill_kill; 
    Gtk2->main_quit
}
